# Copyright (c) OpenMMLab. All rights reserved.
from .generator_discriminator import ResnetGenerator
from .modules import ResidualBlockWithDropout

__all__ = ['ResnetGenerator', 'ResidualBlockWithDropout']
